########################################################################################
##
##    The Network Class
##    The Heavi Network class manages the components and nodes in the circuit. 
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from __future__ import annotations
from enum import Enum
from typing import List, Callable, Literal
from dataclasses import dataclass
from collections import defaultdict
import numpy as np
from .solvers import solve_MNA_RF, solve_MNA_RF_nopgb, solve_MNA_DC, solve_MNA_TRAN, solve_MNA_TRAN_nopgb
from .node import Node, NodeAllocator
from .sparam import Sparameters
from .numeric import SimParam, enforce_simparam, Function, SimParam, Scalar
from .base import (MatSource, BaseComponent, SimulationType, Port, Admittance, Resistor, Inductor, 
                   Capacitor, Impedance, TransmissionLine, NPortS, NPortY, VoltageDC, Diode, VoltageFN)

from loguru import logger
import numba_progress as nbp

#  __   __        __  ___           ___  __  
# /  ` /  \ |\ | /__`  |   /\  |\ |  |  /__` 
# \__, \__/ | \| .__/  |  /~~\ | \|  |  .__/ 
# -------------------------------------------


TEN_POWERS = {
    -12: "p",
    -9: "n",
    -6: "u",
    -3: "m",
    0: "",
    3: "k",
    6: "M",
    9: "T",
    12: "P",
}


TWOPI = np.float64(2 * np.pi)
PI = np.pi
GEN_NULL = Scalar(0)
NPZERO = np.float64(0)

#  __             __   __  
# /  ` |     /\  /__` /__` 
# \__, |___ /~~\ .__/ .__/ 
# -------------------------------------------

    
class Network:
    """ Network class for the Network object.
    This class represents a network of components and nodes. It is used to build and analyze circuits.
    
    Parameters
    ----------
    default_name : str
        The default name for a node.
    node_name_counter_start : int
        The starting index for the node name counter.
    """

    def __init__(self, default_name: str = 'Node', suppress_loadbar: bool = False):
        self.gnd: Node = Node("gnd", _parent=self, _gnd=True)
        
        self.components: list[BaseComponent] = []
        self.sources: list[MatSource] = []
        self.ports: dict[int, Port] = {}
        self.port_counter: int = 1
        
        self.suppress_loadbar: bool = suppress_loadbar
        self._vramp: float = 0

        self._default_node_index: str = default_name
        self._nodes: list[Node] = [self.gnd]
        self._node_library: dict[str, NodeAllocator] = {self._default_node_index: NodeAllocator(self._default_node_index, self)}
        
        self._initialized: bool = False

    def get_vramp(self):
        return self._vramp
    
    def print_components(self) -> None:
        '''Prints an overview of the components in the Network'''
        for comp in self.components:
            print(comp.summary())

    @property
    def node_names(self) -> list[str]:
        '''A list of strings corresponding to each node.'''
        return [n.name for n in self._nodes]
    
    def unlinked_nodes(self) -> list[Node]:
        '''Returns a list of nodes that are not linked to any other nodes.'''
        return [node for node in self._nodes if node._linked is None]
    
    def _define_indices(self, simtype: SimulationType) -> None:
        '''_define_indices writes an index number to the node's index field required for matrix lookup.
        
        This method is called before running the analysis to ensure that all nodes have an index number.
        '''

        i = 0
        for node in self._nodes:
            if node._linked is not None:
                continue
            node.set_index(i)
            i += 1
        
        for component in self.components:
            N = component.n_requested_sources(simtype)
            sources = [MatSource(i) for i in range(i, i+N)]
            self.sources.extend(sources)
            component.set_sources(sources)
            i += N

    
    def _assign_internal_nodes(self, simtype: SimulationType) -> None:
        '''Assigns internal nodes to components that require them.'''
        for component in self.components:
            N = component.n_requested_internal_nodes(simtype)
            nodes = [self.node(index='core_internal') for _ in range(N)]
            component.set_internal_nodes(nodes)
    
    def port(self, number: int) -> BaseComponent | None:
        """Returns the port object corresponding to the provided number.

        Args:
            number (int): The port number to retrieve.

        Returns:
            BaseComponent | None: The port object corresponding to the provided number.
        """        
        return self.ports.get(number, None)

    def get_port_index(self) -> int:
        """Returns the next available port index.

        Returns:
            int: The next available port index.
        """        
        index = self.port_counter
        self.port_counter += 1
        return index
    
    def node(self, number: int = None, index: str = None) -> Node:
        """Return the node with the provided number from the provided node index library.

        Args:
            number (int, optional): The noden umber. Defaults to None.
            index (str, optional): The index. Defaults to None. (default index)

        Returns:
            Node: The node object
        """        
        if index is None:
            index = self._default_node_index
        if index not in self._node_library:
            self._node_library[index] = NodeAllocator(index, self)
        
        node, is_new = self._node_library[index].request(number)
        if is_new:
            self._nodes.append(node)
        return node


    def mnodes(self, N: int, name: str = None) -> list[Node]:
        """Creates a list of N nodes in the network.

        Args:
            N (int): The number of nodes to create.
            name (str, optional): The node name prefix. Defaults to None.

        Returns:
            list[Node]: A list of N nodes.
        """        
        if name is None:
            return [self.node() for _ in range(N)]
        else:
            return [self.node(f'{name}_{i+1}') for i in range(N)]
    
    def _check_unconnected_nodes(self) -> None:
        '''Checks for unconnected nodes in the network and raises a warning if any are found.'''
        # collecct a list of nodes and included status
        node_dict = {node.unique(): False for node in self._nodes}

        # check if nodes are used in components
        for component in self.components:
            for node in component.all_nodes:
                node_dict[node.unique()] = True
        
        for node, used in node_dict.items():
            if not used:
                logger.error(f"Node {node.name} is not connected to any components.")
                logger.error("Unconnected nodes will cause the analysis to yield 0 values.")
                raise ValueError(f"Node {node.name} is not connected to any components.")

    def _get_components(self, simtype: SimulationType) -> list[BaseComponent]:
        '''Returns a list of components that are compatible with the simulation type.'''
        component_list = []
        for component in self.components:
            if simtype in component.supported_simulations:
                component_list.append(component)
            else:
                logger.warning(f"Component {component.component_name} does not support {simtype} simulation. Component is ignored")
        return component_list
    
    def run_dc(self, 
               maxiter: int = 100, 
               source_ramp_steps: int = 0,
               voltage_interpolation_steps: int = 201) -> np.ndarray:
        """Runs a DC analysis on the network.

        Args:
            maxiter (int, optional): _description_. Defaults to 100.
            source_ramp_steps (int, optional): _description_. Defaults to 0.
            voltage_interpolation_steps (int, optional): _description_. Defaults to 201.

        Returns:
            np.ndarray: _description_
        """        
        logger.warning("DC Analysis is still under development and will not work for most components.")

        simtype = SimulationType.DC

        self._assign_internal_nodes(simtype)
        self._define_indices(simtype)
        self._check_unconnected_nodes()
        if all([c.is_linear(simtype) for c in self._get_components(simtype)]):
            maxiter = 1
            
        M = len(self.sources)
        N = max([node.index for node in self._nodes]) + 1

        Vsol = np.zeros((N, ))
        
        A_compilers = [c.DC_matrix_compiler() for c in self.components]
        I_compilers = [c.DC_source_compiler() for c in self.components]
        A_compilers = [c for c in A_compilers if c is not None]
        I_compilers = [c for c in I_compilers if c is not None]

        SolVec = np.zeros((M+N,))
        
        if source_ramp_steps == 0:
            rampvals = np.array([1.0,])
        else:
            rampvals = np.linspace(0, 1, source_ramp_steps+2)[1:]


        A = np.zeros((M+N, M+N)) 

        DiodeData = [(-1,-1,np.linspace(0,1,5),np.linspace(0,1,5)),]
        ## 
        for comp in self.components:
            if not isinstance(comp, Diode):
                continue
            I, V = comp.IVData(Nsteps=voltage_interpolation_steps)
            DiodeData.append((comp.nodes[0].index, comp.nodes[1].index, I,V))

        for A_compiler in A_compilers:
            A = A_compiler(A)
        
        for I_compiler in I_compilers:
            SolVec = I_compiler(SolVec)

        Vsol = solve_MNA_DC(A, SolVec, N, M, rampvals, maxiter, DiodeData, 0.001, 0.001) 

        return Vsol
    
    def run_transient(self, duration: float, dt: float, alpha: float = 0.5,
                      maxiter: int = 5, 
                      source_ramp_steps: int = 0,
                      voltage_interpolation_steps: int = 51) -> np.ndarray:
        
        logger.warning("Transient Analysis is still under development and will not work for most components.")

        simtype = SimulationType.TRANS

        self._assign_internal_nodes(simtype)
        self._define_indices(simtype)
        self._check_unconnected_nodes()

        if all([c.is_linear(simtype) for c in self._get_components(simtype)]):
            maxiter = 1
        
        timesteps = np.linspace(0, duration, int(duration/dt)+1)
        n_timesteps = len(timesteps)

        M = len(self.sources)
        N = max([node.index for node in self._nodes]) + 1

        Vsol = np.zeros((N, ))
        
        A_compilers = [c.TRANS_matrix_compiler() for c in self.components]
        I_compilers = [c.TRANS_source_compiler() for c in self.components]
        A_compilers = [c for c in A_compilers if c is not None]
        I_compilers = [c for c in I_compilers if c is not None]

        SolVec = np.zeros((M+N,))
        
        if source_ramp_steps == 0:
            rampvals = np.array([1.0,])
        else:
            rampvals = np.linspace(0, 1, source_ramp_steps+2)[1:]

        A = np.zeros((M+N, M+N)) 

        DiodeData = [(-1,-1,np.linspace(0,1,5),np.linspace(0,1,5)),]
        LCData = [(-1, -1,-1,-1, 0.0),]

        ## GENERATE TIMESERIES
        transient_source_data = [(-1,np.linspace(0,1,5),np.linspace(0,1,5)),]
        for source in self.components:
            if isinstance(source, VoltageFN):
                transient_source_data.append((source.index, source.vfun(timesteps), timesteps))
        
        for comp in self.components:
            if isinstance(comp, Diode):
                I, V = comp.IVData(Nsteps=voltage_interpolation_steps)
                DiodeData.append((comp.nodes[0].index, comp.nodes[1].index, I,V))
            elif isinstance(comp, Capacitor):
                LCData.append((0, comp.nodes[0].index, comp.nodes[1].index, -1, comp.C))
            elif isinstance(comp, Inductor):
                LCData.append((1, comp.nodes[0].index, comp.nodes[1].index, comp.sources[0].index, comp.L))

        for A_compiler in A_compilers:
            A = A_compiler(A)
        
        for I_compiler in I_compilers:
            SolVec = I_compiler(SolVec)

        if self.suppress_loadbar:
            Vsol = solve_MNA_TRAN_nopgb(A, SolVec, N, M, transient_source_data, rampvals, maxiter, DiodeData) 

        else:
            with nbp.ProgressBar(total=n_timesteps) as progress:
                Vsol = solve_MNA_TRAN(A, SolVec, N, M, transient_source_data, rampvals, maxiter, DiodeData, LCData, alpha, progress) 

        
        return timesteps, Vsol
        
    def run_sp(self, frequencies: np.ndarray, reinitialize: bool = False) -> Sparameters:
        """
        Runs an S-parameter analysis using the MNA method for the network at the specified frequencies.

        Parameters:
        -----------
        frequencies (np.ndarray): An array of frequencies at which to run the analysis.

        Returns:
        --------
        Sparameters: An Sparameters object containing the S-parameter matrix for the network at the specified frequencies

        """

        # Prepare the network for simulation

        simtype = SimulationType.SP
        
        if not self._initialized:
            self._assign_internal_nodes(simtype)
            self._define_indices(simtype)
            self._check_unconnected_nodes()
            self._initialized = True

        # Define the number of sources and nodes

        M = len(self.sources)
        nF = len(frequencies)
        N = max([node.index for node in self._nodes]) + 1
        
        # Initialize the S-parameter matrix and voltage source compiler
        A_compilers = [c.SP_matrix_compiler() for c in self._get_components(simtype)]
        A_compilers = [c for c in A_compilers if c is not None]


        total_number_of_computations = M * nF

        # Initialize the S-parameter matrix and voltage source compiler
        mna_matrix = np.zeros((M+N, M+N, nF), dtype=np.complex128)
        for compiler in A_compilers:
            mna_matrix = compiler(mna_matrix, frequencies)

        # Initialize the vector containing all impedances
        impedance_vector = np.zeros((M,nF), dtype=np.complex128)

        # Initialize the vector containing all S-parameter port indices.
        voltage_source_nodes = []
        for iport, port in self.ports.items():
            voltage_source_nodes.append((iport-1, port.sig_node.index, port.int_node.index, port.gnd_node.index))
            impedance_vector[iport-1,:] = port.Z0(frequencies)
        indices = np.array(voltage_source_nodes).astype(np.int32)

        # Initialize the frequencies array in the right data format
        frequencies = np.array(frequencies).astype(np.float32)

        Sol = None

        # Run the simulation
        if self.suppress_loadbar:
            V, Sol = solve_MNA_RF_nopgb(mna_matrix, impedance_vector, indices, frequencies, N, M)
        else:
            with nbp.ProgressBar(total=total_number_of_computations) as progress:
                V, Sol = solve_MNA_RF(mna_matrix, impedance_vector, indices, frequencies, N, M, progress) 
        
        if reinitialize:
            self._initialized = False

        return Sparameters(Sol, frequencies)
    

    def DC_source(self, node: Node, voltage: float | SimParam, gnd: Node = None, r_series: float | SimParam = 0.001) -> VoltageDC:
        """Creates a DC voltage source between two nodes in the circuit.

        Args:
            node (Node): The node to which the voltage source is connected.
            voltage (float | SimParam): The voltage value of the source.
            gnd (Node, optional): The negative terminal node. Defaults to None.
            r_series (float | SimParam, optional): The series resistance. Defaults to 0.001.

        Returns:
            VoltageDC: The created DC voltage source component.
        """        
        if gnd is None:
            gnd = self.gnd
        vdc = VoltageDC(gnd, node, voltage, r_series=r_series)
        self.components.append(vdc)
        return vdc
    
    def TRANSIENT_source(self, node: Node,
                         voltage_function: Callable[[float], float],
                         gnd: Node = None,
                         r_series: float | SimParam = 0.001) -> VoltageFN:
        """Creates a transient voltage source between two nodes in the circuit.
        
        Args:
            node (Node): The node to which the voltage source is connected.
            voltage_function (Callable[[float], float]): The voltage function of the source.
            gnd (Node, optional): The negative terminal node. Defaults to None.
            r_series (float | SimParam, optional): The series resistance. Defaults to 0.001.
        
        Returns:
            VoltageFN: The created transient voltage source component.

        """
        if gnd is None:
            gnd = self.gnd
        vfn = VoltageFN(gnd, node, voltage_function, r_series=r_series)
        self.components.append(vfn)
        return vfn
    
    def new_port(self, Z0: float, node: Node = None, gnd: Node = None) -> Port:
        """Creates a new port plus its own output node in the network.

        Args:
            Z0 (float): The characteristic impedance of the port.
            node (Node, optional): The node to which the port is connected. Defaults to None.
            gnd (Node, optional): The reference ground node. Defaults to None.

        Returns:
            Port: The created port object.
        """        
        if gnd is None:
            gnd = self.gnd

        int_node = self.node()

        if node is None:
            src_node = self.node()
        else:
            src_node = node
        port = Port(self.get_port_index(), src_node, int_node, gnd, Z0)
        self.components.append(port)
        self.ports[port.index] = port
        
        return port
    
    def quick_port(self, Z0: float, gnd: Node = None) -> Node:
        """Creates a new port and returns only signal node of the port.

        Args:
            Z0 (float): The characteristic impedance of the port.
            gnd (Node, optional): The reference ground node. Defaults to None.

        Returns:
            Node: The signal node of the created port.
        """        
        port = self.new_port(Z0, gnd)
        return port.sig_node
    
    def admittance(self, node1: Node, node2: Node, Y: float) -> Admittance:
        """Creates and returns a component object for an admittance.

        Args:
            node1 (Node): The first node of the admittance.
            node2 (Node): The second node of the admittance.
            Y (float): The admittance value of the admittance in Siemens.

        Returns:
            Admittance: The created admittance component object.
        """        
        admittance_obj = Admittance(node1, node2, Y)
        self.components.append(admittance_obj)
        return admittance_obj

    def impedance(self, node1: Node, node2: Node, Z: float,
                  display_value: float = None) -> Impedance:
        """Creates and returns a component object for an impedance.

        Parameters:
        -----------
        node1 (Node): The first node of the impedance.
        node2 (Node): The second node of the impedance.
        Z (float): The impedance value of the impedance in ohms.
        component_type (ComponentType, optional): The type of the component. Defaults to ComponentType.IMPEDANCE.
        display_value (float, optional): The value to display for the component. Defaults to None.

        Returns:
        --------
        Impedance: The created impedance component object.

        """
        imp = Impedance(node1, node2, Z)
        self.components.append(imp)
        return imp

    def resistor(self, node1: Node, node2: Node, R: float):
        """
        Adds a resistor between two nodes in the circuit.

        Parameters:
        -----------
            node1 (Node): The first node to which the resistor is connected.
            node2 (Node): The second node to which the resistor is connected.
            R (float): The resistance value of the resistor in ohms.

        Returns:
        --------
        Impedance: The impedance object representing the resistor between the two nodes.
        """
        res = Resistor(node1, node2, R)
        self.components.append(res)
        return res
    
    def capacitor(self, node1: Node, node2: Node, C: float) -> Capacitor:
        """
        Creates and returns a component object for a capacitor.

        Parameters:
        -----------

        node1 (Node): The first node of the capacitor.
        node2 (Node): The second node of the capacitor.
        C (float): The capacitance value of the capacitor in Farads.

        Returns:
        --------
        Component: The created capacitor component object.

        """
        cap = Capacitor(node1, node2, C)
        self.components.append(cap)
        return cap
        
    def inductor(self, node1: Node, node2: Node, L: float):
        """
        Adds an inductor component between two nodes in the circuit.
        Args:
            node1 (Node): The first node to which the inductor is connected.
            node2 (Node): The second node to which the inductor is connected.
            L (float): The inductance value of the inductor in Henrys.
        Returns:
            Component: The created inductor component.
        """
        ind = Inductor(node1, node2, L)
        self.components.append(ind)
        return ind
    
    def diode(self, node1: Node, node2: Node,
              Is: float = 1.2e-15, n: float = 1, Vt: float = 0.0258) -> Diode:
        """
        Adds a diode component between two nodes in the circuit.
        Args:
            node1 (Node): The first node to which the diode is connected.
            node2 (Node): The second node to which the diode is connected.
            Is (float): The saturation current of the diode.
            n (float): The emission coefficient of the diode.
            Vt (float): The thermal voltage of the diode.
        Returns:
            Component: The created diode component.
        """
        diode = Diode(node1, node2, Vt, n, Is, 0.15)
        self.components.append(diode)
        return diode
    
    def transmissionline(
        self, port1: Node, port2: Node, Z0: float, er: float, L: float, gnd: Node = None
    ) -> TransmissionLine:
        """
        Creates and returns a component object for a transmission line.

        Parameters:
        -----------
        gnd (Node): The ground node.
        port1 (Node): The first port node.
        port2 (Node): The second port node.
        Z0 (float): Characteristic impedance of the transmission line.
        er (float): Relative permittivity of the transmission line.
        L (float): Length of the transmission line.

        Returns:
        --------
        Component: A component object representing the transmission line.
        """
        c0 = 299792458
        func_er = enforce_simparam(er)
        func_Z0 = enforce_simparam(Z0, unit='Ω')
        def beta(f):
            return TWOPI * f / c0 * np.sqrt(func_er(f))
        
        return self.TL(port1, port2, Function(beta), L, func_Z0, gnd)

    def TL(self, node1: Node, node2: Node, beta: float | SimParam, length: float, Z0: float | SimParam, gnd: Node = None) -> TransmissionLine:
        if gnd is None:
            gnd = self.gnd
        tl = TransmissionLine(node1, node2, gnd, Z0, length, beta)
        self.components.append(tl)
        return tl
    
    def n_port_S(
            self,
            gnd: Node,
            nodes: list[Node],
            Sparam: list[list[Callable]],
            Z0: float,
    ) -> NPortS:
        """Adds an N-port S-parameter component to the circuit.

        Parameters:
        -----------
        gnd : Node
            The ground node of the circuit.
        nodes : list[Node]
            List of nodes representing the ports of the N-port network.
        Sparam : list[list[Callable]]
            A nested list of callables representing the S-parameters as functions of frequency.
        Z0 : float
            The reference impedance.
        Returns:
        --------
        None
        Notes:
        ------
        This method constructs the admittance matrix (Y-parameters) from the given S-parameters
        and adds the corresponding component to the circuit's component list.
        """
        nps = NPortS(nodes, gnd, Sparam, Z0)
        self.components.append(nps)
        return nps

    def n_port_Y(
            self,
            gnd: Node,
            nodes: list[Node],
            Yparams: list[list[Callable]],
            Z0: float,
    ) -> NPortY:
        """Adds an N-port Y-parameter component to the circuit.

        Parameters:
        -----------
        gnd : Node
            The ground node of the circuit.
        nodes : list[Node]
            List of nodes representing the ports of the N-port network.
        Yparam : list[list[Callable]]
            A nested list of callables representing the Y-parameters as functions of frequency.
        Z0 : float
            The reference impedance.
        Returns:
        --------
        None
        Notes:
        ------
        This method constructs the admittance matrix (Y-parameters)and adds the corresponding component 
        to the circuit's component list.
        """
        npy = NPortY(nodes, gnd, Yparams, Z0)
        self.components.append(npy)
        return npy



import numpy as np
import numba as nb

@nb.njit(cache=True)
def stack(matrix: np.ndarray, vector: np.ndarray, add: np.ndarray) -> np.ndarray:
    n1, n2 = matrix.shape
    n3 = vector.shape[0]
    newmat = np.zeros((n1, n2, n3), dtype=np.float64)
    for i in range(n3):
        newmat[:, :, i] = matrix * vector[i] + add
    return newmat

@nb.njit(cache=True)
def cstack(matrix: np.ndarray, vector: np.ndarray, add: np.ndarray) -> np.ndarray:
    n1, n2 = matrix.shape
    n3 = vector.shape[0]
    newmat = np.zeros((n1, n2, n3), dtype=np.complex128)
    for i in range(n3):
        newmat[:, :, i] = matrix * vector[i] + add
    return newmat


########################################################################################
##
##    The SubCircuit class is a base class for all components in the library. 
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------

from __future__ import annotations
from ..network import Network, Node

#  __             __   __   ___  __  
# /  ` |     /\  /__` /__` |__  /__` 
# \__, |___ /~~\ .__/ .__/ |___ .__/ 
# -------------------------------------------


class SubCircuit:
    """ The SubCircuit class is a base class for all components in the library.

    The SubCircuit class can be used as inheritance for defining custom SubCircuits.

    To define a custom SubCircuit, create a new class that inherits from SubCircuit.
    Then Make sure to call the super().__init__() method in the __init__ method of the new class.
    Finally, implement the __on_connect__ method to define the behavior of the SubCircuit.
    The __on_connect__ method is called when all nodes are connected to the network.

    All SubCircuit objects can be connected to nodes using the .connect() method.

    Example:
    --------
    class CustomSubCircuit(SubCircuit):
        def __init__(self):
            super().__init__()
        
        def __on_connect__(self):
            print("All nodes are connected.")

    circuit = CustomSubCircuit()
    circuit.connect(node1, node2, gnd=gnd)

    """     
    def __init__(self):
        self.network: Network = None
        self.nodes: dict[int, Node] = {}
        self.gnd: Node = None
        self.n_nodes: int = 0
        self.__post_init__()

        self._selected_terminal: int = None
    
    def __post_init__(self):
        pass

    @property
    def first_node(self) -> int:
        for i in range(1,self.n_nodes):
            if i not in self.nodes:
                return i
        return self.n_nodes
            
    def __update__(self):
        for node in self.nodes.values():
            self.network = node._parent
        
    def __validate__(self):
        # Check if number of nodes > 0
        if self.n_nodes == 0:
            raise ValueError("The component must have at least one node.")
        
        # Check if self.network is a Network object
        if not isinstance(self.network, Network):
            raise ValueError("The component must be connected to a Network object.")
        
        # Check if all nodes are Node objects with the same Network
        for node in self.nodes.values():
            if not isinstance(node, Node):
                raise ValueError(f"All nodes must be Node objects. Got {type(node)} instead.")
            if node._parent != self.network:
                raise ValueError("All nodes must belong to the same Network object.")
        
        # Check if the ground is defined, if its also a node and its network is the same network.
        if self.gnd is not None:
            if not isinstance(self.gnd, Node):
                raise ValueError("The ground must be a Node object.")
            if self.gnd._parent != self.network:
                raise ValueError("The ground node must belong to the same Network object.")

    def _check_full(self) -> None:
        """ Check if all nodes are connected, and if so, it calls __on_connect__(). """

        for i in range(1, self.n_nodes+1):
            if i not in self.nodes:
                return
        self.__on_connect__()

    def __on_connect__(self):
        raise NotImplementedError("This method must be implemented in the child class.")
    
    def t(self, index: int) -> SubCircuit:
        self._selected_terminal = index
        return self
    
    def __lt__(self, other: Node) -> SubCircuit:
        if isinstance(other, Node):
            self.partial_connect(self._selected_terminal, other)
            return self
        return NotImplemented
    
    def __gt__(self, other: Node) -> SubCircuit:
        if isinstance(other, Node):
            self.partial_connect(self._selected_terminal, other)
            return self
        return NotImplemented
    
    def node(self, index: int) -> Node:
        if index not in self.nodes:
            self.nodes[index] = self.network.node()
        return self.nodes.get(index, None)

    def partial_connect(self, index: int, node: Node) -> SubCircuit:
        self.nodes[index] = node
        self.__update__()
        self._check_full()
        return self
    
    def connect(self, *nodes, gnd: Node = None) -> SubCircuit:
        """ Connect the component to the network. 

        Parameters:
        -----------
        *nodes : Node
            The nodes to connect the component to.
        gnd : Node
            The ground node to connect the component to.
        """
        self.gnd = gnd
        for i, node in enumerate(nodes):
            self.nodes[i+1] = node

        self.__update__()

        if self.gnd is None:
            self.gnd = self.network.gnd

        self.__validate__()
        self.__on_connect__()
        return self


#  __   ___  __          ___  __      __             __   __   ___  __  
# |  \ |__  |__) | \  / |__  |  \    /  ` |     /\  /__` /__` |__  /__` 
# |__/ |___ |  \ |  \/  |___ |__/    \__, |___ /~~\ .__/ .__/ |___ .__/ 
# -------------------------------------------


class TwoNodeSubCircuit(SubCircuit):

    def __post_init__(self):
        self.n_nodes = 2

    def __lt__(self, other: Node) -> SubCircuit:
        if not isinstance(other, Node):
            return NotImplemented
        self.partial_connect(1, other)
        return self

    def __gt__(self, other: Node) -> Node:
        if not isinstance(other, Node):
            return NotImplemented
        self.partial_connect(2, other)
        return other
    
class ThreeNodeSubCircuit(SubCircuit):

    def __post_init__(self):
        self.n_nodes = 3

class FourNodeSubCircuit(SubCircuit):

    def __post_init__(self):
        self.n_nodes = 4


"""
In This example we will look at creating Cauer filters automatically with the Cauer filter designer in Heavi.
"""

import heavi as hv
import numpy as np

np.set_printoptions(precision=2, linewidth=200)

# We again create our model and nodes
model = hv.Model()

n1 = model.quick_port(50)
n2 = model.quick_port(50)

# In the Model class we have some common filter design features.
# We will create a band-pass filter at 2.5GHz with a bandwidth of 0.2GHz. 5th order with 0.05dB Ripple.
model.filters.cauer_filter(model.gnd, n1, n2, 2.5e9, 0.2e9, 5, 0.05, hv.FilterType.CHEBYCHEV, hv.BandType.BANDPASS, 50)

# We can view the components we created like this.
model.print_components()

# And again we simulate
fs = hv.frange(2e9, 3e9, 1_00_001)

S = model.run_sp(fs)

hv.plot_s_parameters(fs, [S.S11, S.S21], labels=['S11', 'S21'])



########################################################################################
##
##    A Touchstone file parser for Heavi
##    A touchstone file importer for S-parameter data.
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from __future__ import annotations
from .libgen import SubCircuit
import numpy as np
from scipy.interpolate import interp1d
from pathlib import Path
from typing import Callable
import re
from loguru import logger

#  __   __        __  ___           ___  __  
# /  ` /  \ |\ | /__`  |   /\  |\ |  |  /__` 
# \__, \__/ | \| .__/  |  /~~\ | \|  |  .__/ 
# -------------------------------------------


_FUNIT = {
    'hz': 1,
    'khz': 1000,
    'mhz': 1e6,
    'ghz': 1e9,
}
_NPORTS = {
    '.s1p': 1,
    '.s2p': 2,
    '.s3p': 3,
    '.s4p': 4,
    '.snp': None
}



#  __   __             ___         ___       __   ___     ___            __  ___    __        __  
# /  ` /  \ |\ | \  / |__  |\ | | |__  |\ | /  ` |__     |__  |  | |\ | /  `  |  | /  \ |\ | /__` 
# \__, \__/ | \|  \/  |___ | \| | |___ | \| \__, |___    |    \__/ | \| \__,  |  | \__/ | \| .__/ 
# -------------------------------------------


def _ma_ri(mag: float, angle: float) -> complex:
    """Converts magnitude and angle to complex number.

    Args:
        mag (float): The magnitude of the complex number.
        angle (float): The angle of the complex number.

    Returns:
        complex: The complex number.
    """    
    return mag * np.exp(1j * np.radians(angle))

def _db_ri(db: float, angle: float) -> complex:
    """Converts dB and angle to complex number.

    Args:
        db (float): The magnitude in dB.
        angle (float): The angle in degrees.

    Returns:
        complex: The complex number.
    """    
    return 10**(db/20) * np.exp(1j * np.radians(angle))

def _ri_ri(real: float, imag: float) -> complex:
    """Converts real and imaginary parts to complex number.

    Args:
        real (float): The real part of the complex number.
        imag (float): The imaginary part of the complex number.

    Returns:
        complex: The complex number.
    """    
    return real + 1j * imag

_DATAMAP = {
    'ma': _ma_ri,
    'db': _db_ri,
    'ri': _ri_ri
}

#  __             __   __   ___  __  
# /  ` |     /\  /__` /__` |__  /__` 
# \__, |___ /~~\ .__/ .__/ |___ .__/ 
# -------------------------------------------


class FileBasedNPort(SubCircuit):
    """A touchstone file importer for S-parameter data."

    The FileBasedNPort class is a subclass of the SubCircuit class and is intended to be used
    to import S-parameter data from touchstone files. The class is intended to be used in Heavi
    simulations and can be used to import S-parameter data from touchstone files.

    Args:
        filename (str): The filename of the touchstone file.
        n_ports (int, optional): The number of ports. Defaults to None.
        ignore_extension (bool, optional): Ignore the file extension. Defaults to False.
        interp_type (str, optional): The interpolation type. Defaults to 'cubic'.

    Example:
        ```python
        touchstone = FileBasedNPort('touchstone.s2p')
        touchstone.renormalize(75)
        touchstone.connect(n1, n2)
        ```

    """
    def __init__(self, filename: str,
                 n_ports: int = None,
                 ignore_extension: bool = False,
                 interp_type: str = 'cubic'):
        super().__init__()
        path = Path(filename)
        if not path.exists():
            raise ValueError(f'The provided filename {path} does not exist.')

        if path.suffix not in ('.s1p','.s2p','.s3p','.s4p','.snp','.ts') and not ignore_extension:
            raise ValueError(f'Can only open touchstone files, files with {path.suffix} are not a valid touchstone extension.')
        
        with open(filename,'r') as f:
            self._lines = f.read().split('\n')
        
        self.Sdata = None

        self.freq_unit: float = 1e9
        self.param_type: str = 's'
        self.data_type: str = 'ma'
        self._converter: Callable = _DATAMAP['ma']
        self.refimp: float = 50
        self.n_ports: int = None
        self.interp_type: str = interp_type
        self.Sdata: np.ndarray = None
        self.Fdata: np.ndarray = None

        if self.param_type != 's':
            raise ValueError('Only S-parameter touchstone files are supported as of this moment.')

        self._parse_touchstone()
    
    def summarize_data(self):
        print(f'Frequency unit: {self.freq_unit}')
        print(f'Parameter type: {self.param_type}')
        print(f'Data type: {self.data_type}')
        print(f'Converter: {self._converter}')
        print(f'Reference impedance: {self.refimp}')
        print(f'Number of ports: {self.n_ports}')

    def _parse_touchstone(self) -> None:
        s_data = []
        f_data = []
        s_collector = []
        freq = None
        for line in self._lines:
            if not line:
                continue
            
            stripped = line.strip().lower()
            # Comment line
            if stripped[0]=='!':
                continue
            
            # Options line
            if stripped[0]=='#':
                funit = re.findall(r'(hz|khz|mhz|ghz)', stripped)
                param_type = re.findall(r'([syzgh])', stripped)
                data_type = re.findall(r'(ma|db|ri)', stripped)
                refimp = re.findall(r'r\s+(\d+)', stripped)

                if funit:
                    self.freq_unit = _FUNIT[funit[0]]
                if param_type:
                    self.param_type = param_type[0]
                if data_type:
                    self.data_type = data_type[0]
                    self._converter = _DATAMAP[data_type[0]]
                if refimp:
                    self.refimp = float(refimp[0])
            
                continue
            
            # Version 2.0 Keyword line:
            if stripped[0]=='[':
                logger.warning('Version 2.0 touchstone files are not supported yet and keywords will be ignored')
                continue
            # Data line
            # remove the exclamation mark plus all symbols behind it if it occurs
            line = line.split('!')[0]

            nums = [float(x) for x in line.split(' ') if x]

            if len(nums)%2 == 1:
                if len(s_collector) > 0:
                    s_data.append(s_collector)
                    f_data.append(freq)
                s_collector = nums[1:]
                freq = nums[0]
            else:
                s_collector.extend(nums)
            
        s_data.append(s_collector)
        f_data.append(freq)

        ssample = s_data[0]
        self.n_ports = int(np.sqrt(len(ssample) // 2))
        self.n_nodes = self.n_ports

        self.n_frequencies = len(f_data)
        self.Fdata = np.array(f_data) * self.freq_unit
        self.Sdata = np.zeros((self.n_frequencies, self.n_ports, self.n_ports), dtype=complex)

        for i, s in enumerate(s_data):
            sdata = [self._converter(s1,s2) for s1,s2 in zip(s[::2], s[1::2])]
            Smat = np.array(sdata).reshape(self.n_ports, self.n_ports)
            if self.n_ports == 2:
                Smat = Smat.T
            self.Sdata[i,:,:] = Smat

    def renormalize(self, new_impedance: float) -> FileBasedNPort:
        """ Renormalize the touchstone file to a new impedance.

        Args:
            new_impedance (float): The new impedance to normalize the touchstone file to.

        Returns:
            FileBasedNPort: The renormalized touchstone file subcircuit.
        """        
        Zold = self.refimp
        Znew = new_impedance
        A = np.eye(self.n_ports, self.n_ports) * np.sqrt(Znew/Zold)*(1/(Znew+Zold))
        R = np.eye(self.n_ports, self.n_ports) * ((Znew-Zold)/(Znew+Zold))
        iA = np.linalg.pinv(A)
        E = np.eye(self.n_ports, self.n_ports)
        for iif in range(self.n_frequencies):
            S = self.Sdata[iif,:,:]
            self.Sdata[iif,:,:] = iA @ (S - R) @ np.linalg.pinv(E - R @ S) @ A
        self.refimp = new_impedance

    def __on_connect__(self):
        s_functions = []
        for i in range(self.n_ports):
            row = []
            for j in range(self.n_ports):
                row.append(interp1d(self.Fdata, self.Sdata[:,i,j], kind=self.interp_type, bounds_error=False, fill_value=(self.Sdata[0,i,j],self.Sdata[-1,i,j])))
            s_functions.append(row)
        

        self.network.n_port_S(self.network.gnd, 
                              [self.nodes[i] for i in range(1,self.n_ports+1)],s_functions,self.refimp)
        



from .libgen import TwoNodeSubCircuit
from ..numeric import Random, SimParam, MonteCarlo
from ..transformations import VSWR_to_S11, Z_to_S11
import numpy as np


class RandomTwoPort(TwoNodeSubCircuit):
    """Generates a random two-port network component with specified parameters.
    Only to be used in S-parameter networks.

    """    
    def __init__(self, 
                 mcsim: MonteCarlo,
                 length: float, 
                 Z0: float,
                 VSWR: float = 1,
                 er: float = 1,
                 attenuation: float = 0,
                 phase_stability: float = 0,
                 amplitude_stability: float = 0,
                 Z0_error: float = 0,
                 nsigma: float = 3,
                 ):
        """ Generates a random two-port network component with specified parameters.

        This class is only intended to be used in S-parameter simulations with Monte Carlo analysis.

        Args:
            mcsim (MonteCarlo): Monte Carlo simulation object.
            length (float): The Electrical length of the two-port network.
            Z0 (float): The characteristic impedance of the two-port network.
            VSWR (float, optional): The limit of the inputs. Defaults to 1.
            er (float, optional): The dielectric constant associated with the electrical length. Defaults to 1.
            attenuation (float, optional): The port-to-port attentuation. Defaults to 0.
            phase_stability (float, optional): The std. of the phase variation per sim. Defaults to 0.
            amplitude_stability (float, optional): The std. of the amplitude variation. Defaults to 0.
            Z0_error (float, optional): The std. of the port input impedance. Defaults to 0.
            nsigma (float, optional): The number of std. caputed with each provided std limit. Defaults to 3.

        """        
        super().__init__()
        self.mc: MonteCarlo = mcsim

        def kz(f):
            return 2*np.pi*f/(299792458)*np.sqrt(er)
        
        A = mcsim.uniform(1, VSWR)
        B = mcsim.uniform(0, 2*np.pi)
        C = mcsim.gaussian(Z0, Z0_error/nsigma)
        D = mcsim.gaussian(0,amplitude_stability/nsigma)
        E = mcsim.gaussian(0,phase_stability/nsigma)
        self.fS11 = lambda f: np.exp(1j*B(f))*VSWR_to_S11(A(f))*Z_to_S11(C(f), Z0)
        self.fS21 = lambda f: (((1-self.fS11(f)**2)**(0.5))*10**(-(attenuation*length)/20) 
                               * (10**(D(f)/20)) 
                               * np.exp(1j*(kz(f)*length+E(f)*np.pi/180)))
        self.fS12 = self.fS21
        self.fS22 = self.fS11
        self.Z0 = Z0
        
    def __on_connect__(self):
        self.network.n_port_S(self.gnd, [self.node(1), self.node(2)], 
                              [[self.fS11, self.fS12], [self.fS21, self.fS22]], 
                              self.Z0)
        




#  ___      __   ___  __           ___      ___           
# |__  \_/ |__) |__  |__) |  |\/| |__  |\ |  |   /\  |    
# |___ / \ |    |___ |  \ |  |  | |___ | \|  |  /~~\ |___ 
# -------------------------------------------

########################################################################################
##
##    PCB Layouting automization features
##    This PCB layouting library is still under development and is not yet fully functional.
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------

from __future__ import annotations
import numpy as np
from enum import Enum
from .libgen import SubCircuit
from ..network import Network, Node
from scipy.optimize import root_scalar
from loguru import logger

#  __   __             ___         ___       __   ___     ___            __  ___    __        __  
# /  ` /  \ |\ | \  / |__  |\ | | |__  |\ | /  ` |__     |__  |  | |\ | /  `  |  | /  \ |\ | /__` 
# \__, \__/ | \|  \/  |___ | \| | |___ | \| \__, |___    |    \__/ | \| \__,  |  | \__/ | \| .__/ 
# -------------------------------------------


u0 = 4*np.pi*1e-7

def coth(x):
    return -np.tan(x + np.pi/2)

def _microstrip_ereff(w: float, h: float, er: float, t: float):
    u = w/h
    T = t/h
    du1 = T/np.pi * np.log(1 + 4*np.exp(1)/(T*coth(6.517*u)**2))
    dur = 1/2 * (1 + 1/np.cosh(er -1 )) * du1
    u = u + dur
    a = 1 + (1/49) * np.log((u**4 + (u/52)**2)/(u**4 + 0.432)) + (1/18.7) * np.log(1 + (u/18.1)**3)
    b = 0.564 * ((er - 0.9) / (er + 3))**0.053
    return (er + 1)/2 + (er - 1)/2 * (1 + 10/u)**(-a*b)

def _microstrip_z0(w: float, h: float, er: float, t: float):
    # compute hammerstand and jensen
    u = w/h
    T = t/h
    du1 = T/np.pi * np.log(1 + 4*np.exp(1)/(T*coth(6.517*u)**2))
    dur = 1/2 * (1 + 1/np.cosh(er -1 )) * du1
    u = u + dur
    ereff = _microstrip_ereff(w, h, er, t)
    fu = 6 + (2 * np.pi - 6) * np.exp(-((30.666/u)**0.7528))
    Z0freespace = 376.73/np.sqrt(ereff)
    return Z0freespace / (2*np.pi) * np.log(fu/u + np.sqrt(1 + (2/u)**2))

def _stripline_ereff(w, h, er, t):
    return er

def _stripline_z0(w, h, er, t):
    Z0 = 60/np.sqrt(er) * np.log((4*(2*h + t))/(0.67*np.pi *(0.8*w + t)))
    return Z0

def _w_from_z0_microstrip(targetZ0: float, h: float, er: float, t: float):
    # Find the width of the microstrip line that gives the target impedance
    # Use _microstrip_z0 to find the impedance of the line by inverting the function using interpolation
    # Use the scipy.optimize.root_scalar function to find the root of the function
    # Return the width of the line

    def f(w):
        return _microstrip_z0(w, h, er, t) - targetZ0
    
    return root_scalar(f, bracket=[0.00001, 20], xtol=0.000001, method='brentq').root

def _w_from_z0_stripline(targetZ0: float, h: float, er: float, t: float):
    # Find the width of the stripline that gives the target impedance
    # Use _stripline_z0 to find the impedance of the line by inverting the function using interpolation
    # Use the scipy.optimize.root_scalar function to find the root of the function
    # Return the width of the line

    def f(w):
        return _stripline_z0(w, h, er, t) - targetZ0
    
    return root_scalar(f, bracket=[0.00001, 20], xtol=0.000001, method='brentq').root

def _stripline_wz0(height: float, er: float, Z0: float = None, width: float = None) -> tuple[float, float]:
    # Returns both the width and impedance if one is defined
    # Raises an error if none are defined
    if Z0 is None and width is None:
        raise ValueError("Either the width or the impedance must be defined")
    if Z0 is not None:
        width = _w_from_z0_stripline(Z0, height, er)
    if width is not None:
        Z0 = _stripline_z0(width, height, er)
    return width, Z0

def _microstrip_wz0(height: float, er: float, Z0: float = None, width: float = None) -> tuple[float, float]:
    # Returns both the width and impedance if one is defined
    # Raises an error if none are defined
    if Z0 is None and width is None:
        raise ValueError("Either the width or the impedance must be defined")
    if Z0 is not None:
        width = _w_from_z0_microstrip(Z0, height, er)
    if width is not None:
        Z0 = _microstrip_z0(width, height, er)
    return width, Z0

class MicrostripLine(SubCircuit):

    def __init__(self, width: float, length: float, height: float, epsilon_r: float, tand: float, trace_thickenss: float = 60e-6, gnd: Node = None):
        super().__init__()
        self.width: float = width
        self.length: float = length
        self.height: float = height
        self.ereff: float = _microstrip_ereff(width, height, epsilon_r, trace_thickenss)
        self.Z0: float = _microstrip_z0(width, height, epsilon_r, trace_thickenss)
        self.tand: float = tand
        self.gnd: Node = gnd
        self.n_nodes = 2
        

    def __on_connect__(self):
        if self.gnd is None:
            self.gnd = self.network.gnd
        self.network.transmissionline(self.gnd, self.node(1), self.node(2),  self.Z0, self.ereff*(1-1j*self.tand), self.length)

class Stripline(SubCircuit):

    def __init__(self, width: float, length: float, height: float, epsilon_r: float, tand: float, trace_thickenss: float = 60e-6, gnd: Node = None):
        super().__init__()
        self.width: float = width
        self.length: float = length
        self.height: float = height
        self.ereff: float = _stripline_ereff(width, height, epsilon_r, trace_thickenss)
        self.Z0: float = _stripline_z0(width, height, epsilon_r, trace_thickenss)
        self.tand: float = tand
        self.gnd: Node = gnd
        self.n_nodes = 2

    def __on_connect__(self):
        if self.gnd is None:
            self.gnd = self.network.gnd
        self.network.transmissionline(self.gnd, self.node(1), self.node(2),  self.Z0, self.ereff*(1-1j*self.tand), self.length)

class Path:

    def __init__(self, start_node: Node):
        self.start_node = start_node
        self.component_link: list = []
    
    @property
    def last(self):
        return self.component_link[-1]

    @property
    def first(self):
        return len(self.component_link)


class PCBStack:

    def __init__(self, network: Network, epsilon_r: float, tand: float, thickness: float, Nlayers: int):

        # Test that the layers are at least 2 and at most 10
        if Nlayers < 2 or Nlayers > 10:
            raise ValueError("The number of layers must be at least 2 and at most 10")
        
        # Test that the thickness is positive
        if thickness <= 0:
            raise ValueError("The thickness must be positive")
        
        # Test that the relative permittivity is positive
        if epsilon_r <= 0:
            raise ValueError("The relative permittivity must be positive")
        
        # Test that the loss tangent is between 0 and 1
        if tand < 0 or tand > 1:
            raise ValueError("The loss tangent must be between 0 and 1")
        
        self.network: Network = network
        self.epsilon_r: float = epsilon_r
        self.tand: float = tand
        self.thickness: float = thickness
        self.Nlayers: int = Nlayers
        self.trace_thickness: float = 60e-6
    
    @property
    def th(self) -> float:
        return self.thickness/(self.Nlayers-1)
    
    def get_ereff(self, Z0: float, layer: int) -> float:
        if self.is_microstrip(layer):
            width = _w_from_z0_microstrip(Z0, self.th, self.epsilon_r, self.trace_thickness)
            return _microstrip_ereff(width, self.th, self.epsilon_r, self.trace_thickness)
        else:
            width = _w_from_z0_stripline(Z0, self.th, self.epsilon_r, self.trace_thickness)
            return _stripline_ereff(width, self.th, self.epsilon_r, self.trace_thickness)
        
    def is_microstrip(self, layer: int) -> bool:
        return layer == 0 or layer == self.Nlayers - 1
    
    def port(self, impedance: float, layer: int = 0) -> NodeForwarder:
        """ Create a new path starting from a port. """
        n = self.network.new_port(impedance)
        return NodeForwarder(n, layer, self, impedance)

    def transmission_line(self, n1: Node, n2: Node, length: float, impedance: float = None, width: float = None, layer: int = 0, gnd: Node = None) -> NodeForwarder:
        if layer == 0 or layer == self.Nlayers - 1:
            tl = self.microstrip_line(n1, n2, length, impedance=impedance, width=width, gnd=gnd)
        else:
            tl = self.stripline(n1, n2, length, impedance=impedance, width=width, gnd=gnd)

        Z0 = tl.Z0
        if impedance is not None:
            Z0 = impedance
        return NodeForwarder(n2, layer, self, Z0)
    
    def microstrip_line(self, n1: Node, n2: Node, length: float, impedance: float = None, width: float = None, gnd: Node = None) -> MicrostripLine:
        if gnd is None:
            gnd = self.network.gnd
        if impedance is not None and width is None:
            width = _w_from_z0_microstrip(impedance, self.th, self.epsilon_r, self.trace_thickness)
        logger.info(f'Creating microstrip line with width {width} and impedance {impedance}')
        return MicrostripLine(width, length, self.th, self.epsilon_r, self.tand, gnd=gnd).connect(n1, n2)

    def stripline(self, n1: Node, n2: Node, length: float, impedance: float = None, width: float = None, gnd: Node = None) -> Stripline:
        if gnd is None:
            gnd = self.network.gnd
        if impedance is not None:
            width = _w_from_z0_stripline(impedance, self.th, self.epsilon_r, self.trace_thickness)
        return Stripline(width, length, self.th, self.epsilon_r, self.tand, gnd=gnd).connect(n1, n2)
    
class NodeForwarder:

    def __init__(self, node: Node, layer: int, pcb: PCBStack, impedance: float):
        self.node: Node = node
        self.layer: int = layer
        self.pcb: PCBStack = pcb
        self.impedance: float = impedance

    def terminate(self, Z0: float) -> Node:
        """ Terminate the path with a resistor. """
        return self.pcb.network.terminal(self.node, Z0).output_node
    
    def straight(self, L: float, Z0: float = None, width: float = None) -> NodeForwarder:
        """ Create a straight path. """
        if Z0 is None and width is None:
            Z0 = self.impedance
        newnode = self.pcb.network.node()
        return self.pcb.transmission_line(self.node, newnode, L, impedance=Z0, width=width, layer=self.layer)
    
    def capacitor(self, C: float, node: Node = None) -> NodeForwarder:
        """ Create a capacitor. """
        if node is None:
            node = self.pcb.network.node()
            self.pcb.network.capacitor(self.node, node, C)
            return NodeForwarder(node, self.layer, self.pcb, self.impedance)
        else:
            self.pcb.network.capacitor(self.node, node, C)
            return self
        
    def inductor(self, L: float, node: Node = None) -> NodeForwarder:
        """ Create an inductor. """
        if node is None:
            node = self.pcb.network.node()
            self.pcb.network.inductor(self.node, node, L)
            return NodeForwarder(node, self.layer, self.pcb, self.impedance)
        else:
            self.pcb.network.inductor(self.node, node, L)
            return self
        
    def component(self, comp: SubCircuit) -> NodeForwarder | MultiNodeForwarder:
        """ Connect a component to the network. """
        new_nodes = [self.pcb.network.node() for _ in range(comp.n_nodes - 1)]
        comp.connect(self.node, *new_nodes)

        return MultiNodeForwarder([NodeForwarder(node, self.layer, self.pcb, self.impedance) for node in new_nodes])
    
    def via(self, newlayer: int, radius: float = 0.5e-3) -> NodeForwarder:
        """ Create a via to another layer. """
        newnode = self.pcb.network.node()
        n_layer_jump = abs(newlayer - self.layer)
        h = self.pcb.th * n_layer_jump
        r = radius
        ## Use approximal formula for via inductance
        via_inductance = u0/(2*np.pi) *(h*np.log((h+np.sqrt(r**2+h**2))/r))+3/2*(r-np.sqrt(r-np.sqrt(r**2+h**2)))
        self.pcb.network.inductor(self.node, newnode, via_inductance)
        return NodeForwarder(newnode, newlayer, self.pcb, self.impedance)
    
    def Z0_step(self, new_Z0: float) -> NodeForwarder:
        Z1 = self.impedance
        Z2 = new_Z0
        er = self.pcb.epsilon_r
        h = self.pcb.th
        if self.pcb.is_microstrip(self.layer):
            ereff1 = _microstrip_ereff(Z1, self.pcb.th, self.pcb.epsilon_r, self.pcb.trace_thickness)
            w1 = _w_from_z0_microstrip(Z1, self.pcb.th, self.pcb.epsilon_r, self.pcb.trace_thickness)
            ereff2 = _microstrip_ereff(Z2, self.pcb.th, self.pcb.epsilon_r, self.pcb.trace_thickness)
            w2 = _w_from_z0_microstrip(Z2, self.pcb.th, self.pcb.epsilon_r, self.pcb.trace_thickness)
        else:
            ereff1 = _stripline_ereff(Z1, self.pcb.th, self.pcb.epsilon_r, self.pcb.trace_thickness)
            w1 = _w_from_z0_stripline(Z1, self.pcb.th, self.pcb.epsilon_r, self.pcb.trace_thickness)
            ereff2 = _stripline_ereff(Z2, self.pcb.th, self.pcb.epsilon_r, self.pcb.trace_thickness)
            w2 = _w_from_z0_stripline(Z2, self.pcb.th, self.pcb.epsilon_r, self.pcb.trace_thickness)
        
        if w2/w1 < 3.5:
            Cp = (np.sqrt(w1*w2)*(10.1*np.log(er) + 2.33)*w2/w1 - 12.6*np.log(er) - 3.17)*1e-12
        else:
            Cp = (130*np.log(w2/w1)-44)*1e-12

        Lp = h*(40.5*(w1/w2-1.0) - 75*w1/w2 + 0.2*((w1/w2) - 1.0)**2)*1e-9

        n1 = self.pcb.network.node()
        n2 = self.pcb.network.node()

        self.pcb.network.inductor(self.node, n1, Lp)
        self.pcb.network.capacitor(n1, self.pcb.network.gnd, Cp)
        self.pcb.network.inductor(n1, n2, Lp)
        logger.info(f'Adding parasitic inductor {Lp} nH and capacitor {Cp} pF to the Z0 step')

        return NodeForwarder(n2, self.layer, self.pcb, new_Z0)
    
    def right_turn(self, Z0: float = None, miter: bool = False) -> NodeForwarder:
        """ Create a right turn. """
        if Z0 is None:
            Z0 = self.impedance

        n1 = self.pcb.network.node()
        n2 = self.pcb.network.node()
        if self.pcb.is_microstrip(self.layer):
            ereff = _microstrip_ereff(Z0, self.pcb.th, self.pcb.epsilon_r, self.pcb.tand)
            w = _w_from_z0_microstrip(Z0, self.pcb.th, self.pcb.epsilon_r, self.pcb.tand)
        else:
            ereff = _stripline_ereff(Z0, self.pcb.th, self.pcb.epsilon_r, self.pcb.tand)
            w = _w_from_z0_stripline(Z0, self.pcb.th, self.pcb.epsilon_r, self.pcb.tand)
        h = self.pcb.th
        if w/h < 1:
            Cb = w*1e-12*((14*ereff + 12.5)*w/h - (1.83*ereff -2.25))/np.sqrt(w/h)
        else:
            Cb = w((9.5*ereff + 1.25)*(w/h)+5.2*ereff + 7.0)*1e-12
        Lb = h*1e-9*100*np.clip((4*np.sqrt(w/h)-4.21),a_min=0.001, a_max=None)
        logger.info(f'Adding parasitic inductor {Lb} nH and capacitor {Cb} pF to the right turn')
        if not miter:
            self.pcb.network.capacitor(self.node, self.pcb.network.gnd, Cb)
        self.pcb.network.inductor(self.node, n1, Lb)
        self.pcb.network.inductor(n1, n2, Lb)
        return NodeForwarder(n1, self.layer, self.pcb, Z0)
    
    def left_turn(self, Z0: float = None, miter: bool = False) -> NodeForwarder:
        return self.right_turn(Z0)
    
    def chebychev_transformer(self, Ztarget: float, fc: float, Nsections: int = 1, ripple: float = 0.05) -> NodeForwarder:
        from heavi.filtering import impedance_transformer_cheb
        Z1 = self.impedance
        Z2 = Ztarget

        Z0s = impedance_transformer_cheb(Z1, Z2, ripple, Nsections)
        logger.info(f'Adding {Nsections} Chebychev sections with ripple {ripple} to the path with impedances: {Z0s}')
        l0 = 299792458/fc * 0.25

        lastnode = self
        for Z1, Z2 in zip(Z0s[:-1], Z0s[1:-1]):
            L = l0 * 1/(np.sqrt(self.pcb.get_ereff(Z2, self.layer)))
            logger.info(f'Jump from {Z1} to {Z2} with length {L}') 
            lastnode = lastnode.Z0_step(Z2).straight(L,Z2)
        return lastnode.Z0_step(Z0s[-1])
            
class MultiNodeForwarder:

    def __init__(self, nfs: list[NodeForwarder]):
        self.nfs: list[NodeForwarder] = nfs

    def __call__(self, index: int) -> NodeForwarder:
        # check the integer range
        if index < 0 or index >= len(self.nfs):
            raise ValueError("The index must be between 0 and the number of nodes - 1")
        return self.nfs[index]
    


from .component import BaseComponent, Node, SimParam, enforce_simparam, SimulationType, _TWO_MAT
from typing import List, Callable
import numpy as np
from .numeric import stack, cstack



class Admittance(BaseComponent):
    
    supported_simulations: List[SimulationType] = [SimulationType.AC,
                                                   SimulationType.DC, 
                                                   SimulationType.SP,]
    
    def __init__(self, node1: Node, node2: Node, admittance: float | SimParam):
        '''
        Adds a generic admittance component to the circuit. The admittance can have a float
        or frequency dependent value through the SimParameter class.

        Returns:
            Admittance: The admittance component 
        '''
        super().__init__()
        self.nodes = [node1, node2]
        self.nterminals = 2
        self.Y: SimParam = enforce_simparam(admittance)
        self.component_name = 'Admittance'
        self.unit = 'S'
        self.value = self.Y(self.PRINT_FREQ)
    
    @property
    def mat_slice(self) -> tuple[np.ndarray, np.ndarray]:
        ids = np.array([self.nodes[0].index, self.nodes[1].index])
        return np.ix_(ids, ids)
    
    def generic_matrix_compiler(self) -> Callable:
        row,col = self.mat_slice

        def compiler(matrix: np.ndarray, f: float) -> np.ndarray:
            matrix[row,col,:] += cstack(_TWO_MAT, self.Y(f), 0*_TWO_MAT)
            return matrix
        
        return compiler
    
    def SP_matrix_compiler(self):
        return self.generic_matrix_compiler()
    
    def DC_matrix_compiler(self):
        row,col = self.mat_slice

        def compiler(matrix: np.ndarray) -> np.ndarray:
            matrix[row,col] += self.Y(0.0) * _TWO_MAT
            return matrix
        
        return compiler
    
    def AC_matrix_compiler(self):
        return self.generic_matrix_compiler()




from .component import BaseComponent, Node, SimParam, enforce_simparam, SimulationType, _TWO_MAT
from typing import List, Callable
from scipy.optimize import fsolve
from loguru import logger
import numpy as np

class Diode(BaseComponent):
    supported_simulations: List[SimulationType] = [SimulationType.DC, 
                                                   SimulationType.TRANS]
    
    _is_linear = False

    def __init__(self, node1: Node, node2: Node, Vt: float=0.0258, N=1.3497, I0: float = 1.2e-15, Rs = 0.13668):
        super().__init__()
        self.nodes = [node1, node2]
        self.nterminals = 2
        
        self.Vt = Vt
        self.N = N
        self.I0 = I0
        self.Rs = Rs
        
        self.component_name = 'Diode'
        self.unit = 'V'
        self.value = self.Vt
    
    @property
    def mat_slice(self) -> tuple[np.ndarray, np.ndarray]:
        ids = np.array([self.nodes[0].index, self.nodes[1].index])
        return np.ix_(ids, ids)
    
    
    def IVData(self, Vmin: float = -100, Vmax: float = 5, Nsteps: int = 101):
        Vs = np.tan(np.linspace(np.arctan(Vmin), np.arctan(Vmax), Nsteps))
        
        Is = self.I0 * (np.exp(Vs/(self.N*self.Vt))-1)

        return Is, Vs




########################################################################################
##
##    An SMD Subcircuit Library for Heavi
##    This library was written for an old version of Heavi and needs to be revisited.
##    Use at own risk. Intended mostly for S-parameter analysis.
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------


from .libgen import SubCircuit, TwoNodeSubCircuit
from enum import Enum

#  ___                  __  
# |__  |\ | |  |  |\/| /__` 
# |___ | \| \__/  |  | .__/ 
# -------------------------------------------

class SMDResistorSize(Enum):
    R0402 = "0402"
    R0603 = "0603"
    R0805 = "0805"
    R1206 = "1206"
    # Add other sizes as needed

class SMDCapacitorSize(Enum):
    C0402 = "0402"
    C0603 = "0603"
    C0805 = "0805"
    C1206 = "1206"
    # Add other sizes as needed

class SMDInductorSize(Enum):
    L0402 = "0402"
    L0603 = "0603"
    L0805 = "0805"
    L1206 = "1206"
    # Add other sizes as needed

#  __             __   __   ___  __  
# /  ` |     /\  /__` /__` |__  /__` 
# \__, |___ /~~\ .__/ .__/ |___ .__/ 
# -------------------------------------------


class SMDResistor(TwoNodeSubCircuit):
    """
    A simple SMD resistor model that accounts for parasitic inductance and capacitance 
    based on package size.

    Attributes
    ----------
    resistance : float
        The resistance value in ohms.
    inductance : float
        The parasitic inductance in henries.
    capacitance : float
        The parasitic capacitance in farads.
    """

    def __init__(self, resistance: float, package: SMDResistorSize):
        """
        Parameters
        ----------
        resistance : float
            The nominal resistance value (in ohms).
        package : SMDResistorSize
            The package size of the SMD resistor (e.g., 0402, 0603, etc.).
        """
        super().__init__()
        self.n_nodes = 2
        self.resistance = resistance
        self.package = package
        # Typical empirical parasitic values (approximate):
        # These can vary depending on manufacturer, pad layout, etc.
        # You can refine or expand as necessary.
        package_parasitics = {
            SMDResistorSize.R0402: (0.8e-9, 0.2e-12),  # (L, C)
            SMDResistorSize.R0603: (1.0e-9, 0.3e-12),
            SMDResistorSize.R0805: (1.5e-9, 0.45e-12),
            SMDResistorSize.R1206: (1.8e-9, 0.6e-12),
        }
        
        # Fallback for unknown sizes (or define a default)
        default_inductance = 1.0e-9
        default_capacitance = 0.3e-12
        
        if package in package_parasitics:
            self.inductance, self.capacitance = package_parasitics[package]
        else:
            self.inductance = default_inductance
            self.capacitance = default_capacitance

        self.function = None

    def __on_connect__(self):
        """
        Connects the SMD resistor with its parasitic R, L, and C
        into the network. This method is automatically called
        when the component is attached to the network.
        """
        
        def z_parasitic(f):
            # Z_RLC = R + j*2*pi*f*L + 1/(j*2*pi*f*C)
            w = 2 * 3.141592653589793 * f
            Zr = self.resistance
            Zl = 1j * w * self.inductance
            Zc = 1 / (1j * w * self.capacitance)
            # return Zr + Zl parallel to Zc
            return (Zr+Zl) * Zc / (Zr + Zl + Zc)
        self.function = z_parasitic
        self.network.impedance(self.node(1), self.node(2), z_parasitic, display_value=self.resistance)\
        .set_metadata(name=f'{self.package}SMD Resistor',
                      unit='Ω',
                      value=self.resistance,
                      inductance=self.inductance,
                      capacitance=self.capacitance)

class SMDInductor(TwoNodeSubCircuit):
    """
    An SMD inductor with a simple R-L // C parasitic model.
    
    - The inductor has a series ESR (Equivalent Series Resistance) with the inductance.
    - A small parallel capacitance accounts for high-frequency behavior.
    
    Attributes
    ----------
    inductance : float
        The nominal inductance value in henries (H).
    esr : float
        The equivalent series resistance in ohms (Ω).
    parallel_capacitance : float
        The parasitic parallel capacitance in farads (F).
    """

    def __init__(self, inductance: float, package: SMDInductorSize):
        """
        Parameters
        ----------
        inductance : float
            The inductance value in henries.
        package : SMDInductorSize
            The package size of the SMD inductor (e.g. 0402, 0603, etc.).
        """
        super().__init__()
        self.package = package
        self.n_nodes = 2
        self.inductance = inductance
        
        # Typical parasitic values (approximate)
        # Key: SMDInductorSize, Value: (ESR, parallel_capacitance)
        # These are purely illustrative; real values depend on manufacturer & frequency range.
        package_parasitics = {
            SMDInductorSize.L0402: (0.15, 0.3e-12),
            SMDInductorSize.L0603: (0.10, 0.5e-12),
            SMDInductorSize.L0805: (0.08, 0.8e-12),
            SMDInductorSize.L1206: (0.06, 1.2e-12),
        }
        
        # Fallback/default values if package not in dict
        default_esr = 0.1
        default_cpar = 0.5e-12
        
        if package in package_parasitics:
            self.esr, self.parallel_capacitance = package_parasitics[package]
        else:
            self.esr = default_esr
            self.parallel_capacitance = default_cpar

        self.function = None

    def __on_connect__(self):
        """
        Connects the SMD inductor in the network as a single
        frequency-dependent impedance that models:
        
        Z_total = (ESR + jωL) ∥ (1 / jωCpar)
        
        where:
        - ESR is the inductor's equivalent series resistance,
        - L is the nominal inductance,
        - Cpar is the parasitic parallel capacitance.
        """
        import cmath
        import math

        def z_parasitic(f):
            w = 2 * math.pi * f
            
            # Series branch = ESR + jωL
            z_series = self.esr + 1j * w * self.inductance
            
            # Parallel capacitor = 1 / (jωC)
            z_parallel_cap = 1 / (1j * w * self.parallel_capacitance)
            
            # Combine them in parallel:
            # Z_total = (z_series * z_parallel_cap) / (z_series + z_parallel_cap)
            return 1 / (1/(z_series) + 1/(z_parallel_cap))#(z_series * z_parallel_cap) / (z_series + z_parallel_cap)
        self.function = z_parasitic
        # Create the impedance component in the network
        self.network.impedance(self.node(1), self.node(2), z_parasitic, display_value=self.inductance)\
        .set_metadata(name=f'{self.package}SMD Inductor',
                      unit='H',
                      value=self.inductance,
                      esr=self.esr,
                      parallel_capacitance=self.parallel_capacitance)

class SMDCapacitor(TwoNodeSubCircuit):
    """
    An SMD capacitor with an ESL-ESR-C series model.
    
    Attributes
    ----------
    capacitance : float
        The nominal capacitance value in farads (F).
    esr : float
        The equivalent series resistance in ohms (Ω).
    esl : float
        The equivalent series inductance in henries (H).
    """

    def __init__(self, capacitance: float, package: SMDCapacitorSize):
        """
        Parameters
        ----------
        capacitance : float
            The capacitance value in farads.
        package : SMDCapacitorSize
            The package size of the SMD capacitor (e.g., 0402, 0603, etc.).
        """
        super().__init__()
        self.package = package
        self.n_nodes = 2
        self.capacitance = capacitance
        
        # Typical parasitic values (approximate)
        # Key: SMDCapacitorSize, Value: (ESR, ESL)
        package_parasitics = {
            SMDCapacitorSize.C0402: (0.1, 0.3e-9),
            SMDCapacitorSize.C0603: (0.08, 0.5e-9),
            SMDCapacitorSize.C0805: (0.07, 0.7e-9),
            SMDCapacitorSize.C1206: (0.05, 1.0e-9),
        }
        
        # Fallback/default values if package not in dict
        default_esr = 0.1
        default_esl = 0.5e-9
        
        if package in package_parasitics:
            self.esr, self.esl = package_parasitics[package]
        else:
            self.esr = default_esr
            self.esl = default_esl

        self.function = None

    def __on_connect__(self):
        """
        Connects the SMD capacitor in the network as a single
        frequency-dependent impedance that models:
        
        Z_total = ESR + jωESL + 1 / (jωC)
        
        (all components in series).
        """
        import cmath
        import math
        
        def z_parasitic(f):
            w = 2 * math.pi * f
            
            # Series components: ESR + jωESL
            z_series = self.esr + 1j * w * self.esl
            
            # Capacitive reactance = 1 / (jωC)
            z_c = 1 / (1j * w * self.capacitance)
            
            # Total series: Z_total = z_series + z_c
            return z_series + z_c
        self.function = z_parasitic
        # Create the impedance component in the network
        self.network.impedance(self.node(1), self.node(2), z_parasitic, display_value=self.capacitance)\
        .set_metadata(name=f'{self.package}SMD Capacitor',
                      unit='F',
                      value=self.capacitance,
                      esr=self.esr,
                      esl=self.esl)


from .component import BaseComponent, Node, SimParam, enforce_simparam, SimulationType, _TWO_MAT
from typing import List, Callable
import numpy as np

class TransmissionLine(BaseComponent):
    supported_simulations: List[SimulationType] = [SimulationType.SP,]
    
    def __init__(self, node1: Node, node2: Node, gnd: Node, Z0: float | SimParam, length: float, beta: Callable | SimParam):
        super().__init__()
        self.nodes = [node1, node2, gnd]
        self.gnd = gnd
        self.nterminals = 4
        self.Z0: SimParam = enforce_simparam(Z0)
        self.beta: SimParam = enforce_simparam(beta)
        self.length: SimParam = enforce_simparam(length)
        self.component_name = 'TransmissionLine'
        self.unit = 'Ohm'
        self.value = self.Z0.value
    
    @property
    def mat_slice(self) -> tuple[np.ndarray, np.ndarray]:
        ids = np.array([self.nodes[0].index, self.nodes[1].index, self.gnd.index])
        return np.ix_(ids, ids)
    
    def SP_matrix_compiler(self):
        beta = self.beta
        length = self.length
        Z0 = self.Z0
        
        def a11(f, length):
            return np.cosh(1j*beta(f)*length)
        def a12(f, length):
            return Z0(f)*np.sinh(1j*beta(f)*length)
        def a21(f, length):
            return 1/Z0(f)*np.sinh(1j*beta(f)*length)
        def a22(f, length):
            return np.cosh(1j*beta(f)*length)
        
        def y11(f, length):
            return a22(f, length)/a12(f, length)
        def y12(f, length):
            return -((a11(f, length)*a22(f, length))-(a12(f, length)*a21(f, length)))/a12(f, length)
        def y21(f, length):
            return -1/a12(f, length)
        def y22(f, length):
            return a11(f, length)/a12(f, length)
        
        functions = [[y11, y12], [y21, y22]]
        N = 2
        def comp_function(f: float):
            nF = f.shape[0]
            Y = np.array([[y(f, self.length.value) for y in row] for row in functions], dtype=np.complex128)
            Y2 = np.zeros((N+1,N+1,nF),dtype=np.complex128)
            Y2[:N,:N,:] = Y
            for i in range(2):
                #Y2[i,i,:] += np.sum(Y[i,:,:],axis=0)
                Y2[i,N,:] = -np.sum(Y[i,:,:],axis=0)
                Y2[N,i,:] = -np.sum(Y[:,i,:],axis=0)
                Y2[N,N,:] += np.sum(Y[i,:,:],axis=0)
            return Y2
        
        row, col = self.mat_slice
        def compiler(matrix: np.ndarray, f: float) -> np.ndarray:
            matrix[row,col,:] += comp_function(f)
            return matrix
        
        return compiler


from .port import Port
from .admittance import Admittance
from .resistor import Resistor
from .capacitor import Capacitor
from .inductor import Inductor
from .component import MatSource, BaseComponent, SimulationType
from .impedance import Impedance
from .tl import TransmissionLine
from .nport import NPortS, NPortY
from .source import VoltageDC, VoltageFN
from .diode import Diode


########################################################################################
##
##    Commonly used transformations in RF engineering
##    This module contains functions to convert between S11, VSWR and impedance
##
##    Author: Robert Fennis
##    Date: 2025
##
########################################################################################

#          __   __   __  ___  __  
# |  |\/| |__) /  \ |__)  |  /__` 
# |  |  | |    \__/ |  \  |  .__/ 
# -------------------------------------------

import numpy as np

#  ___            __  ___    __        __  
# |__  |  | |\ | /  `  |  | /  \ |\ | /__` 
# |    \__/ | \| \__,  |  | \__/ | \| .__/ 
# -------------------------------------------


def S11_to_VSWR(S11: np.ndarray) -> np.ndarray:
    """Convert S11 to VSWR

    Args:
        S11 (np.ndarray): The reflection coefficient

    Returns:
        np.ndarray: The Voltage Standing Wave Ratio
    """    
    return (1 + np.abs(S11)) / (1 - np.abs(S11))

def S11_to_Z(S11: np.ndarray, Z0: float = 50.) -> np.ndarray:
    """ Convert S11 to impedance

    Args:
        S11 (np.ndarray): The reflection coefficient
        Z0 (float, optional): The reference impedance. Defaults to 50.

    Returns:
        np.ndarray: The aparant impedance
    """    
    return Z0 * (1 + S11) / (1 - S11)

def Z_to_S11(Z: np.ndarray, Z0: float = 50) -> np.ndarray:
    """ Convert impedance to S11

    Args:
        Z (np.ndarray): The load impedance
        Z0 (float, optional): The reference impedance. Defaults to 50.

    Returns:
        np.ndarray: The reflection coefficient
    """    
    return (Z - Z0) / (Z + Z0)

def VSWR_to_S11(VSWR: np.ndarray) -> np.ndarray:
    """Convert VSWR to S11

    Args:
        VSWR (np.ndarray): The Voltage Standing Wave Ratio

    Returns:
        np.ndarray: The reflection coefficient
    """    
    return (VSWR - 1) / (VSWR + 1)

def VSWR_to_Z(VSWR: np.ndarray, Z0: float = 50) -> np.ndarray:
    """Convert VSWR to impedance

    Args:
        VSWR (np.ndarray): The Voltage Standing Wave Ratio
        Z0 (float, optional): The reference impedance. Defaults to 50.

    Returns:
        np.ndarray: The aparant impedance
    """    
    return S11_to_Z(VSWR_to_S11(VSWR), Z0)

def Z_to_VSWR(Z: np.ndarray, Z0: float = 50) -> np.ndarray:
    """Convert impedance to VSWR

    Args:
        Z (np.ndarray): The load impedance
        Z0 (float, optional): The reference impedance. Defaults to 50.

    Returns:
        np.ndarray: The Voltage Standing Wave Ratio
    """    
    return S11_to_VSWR(Z_to_S11(Z, Z0))


from .libgen import SubCircuit, TwoNodeSubCircuit, FourNodeSubCircuit, ThreeNodeSubCircuit
from ..numeric import SimParam, enforce_simparam
from ..network import Node
import numpy as np
from ..filtering import impedance_transformer_cheb

class Wilkinson(ThreeNodeSubCircuit):

    def __init__(self, Z0: float | SimParam, fc: float | SimParam, er: float | SimParam = 1):
        super().__init__()
        self.Z0 = Z0
        self.fc = fc
        self.er = er
    
    def __on_connect__(self):
        wavelength = 299792458 / (self.fc*np.sqrt(self.er))
        def beta(f):
            return 2*np.pi*f*np.sqrt(self.er)/(299792458)
        L = wavelength / 4

        Z0 = self.Z0 * np.sqrt(2)

        self.network.TL(self.node(1), self.node(2), beta, L, Z0)
        self.network.TL(self.node(1), self.node(3), beta, L, Z0)
        self.network.resistor(self.node(2), self.node(3), Z0*2)

class BranchlineCoupler(FourNodeSubCircuit):

    def __init__(self, Z0: float | SimParam, fc: float | SimParam, er: float | SimParam = 1):
        super().__init__()
        self.Z0 = Z0
        self.fc = fc
        self.er = er
    
    def __on_connect__(self):
        wavelength = 299792458 / (self.fc*np.sqrt(self.er))
        L = wavelength / 4
        def beta(f):
            return 2*np.pi*f*np.sqrt(self.er)/(299792458)
        Z0 = self.Z0/np.sqrt(2)

        self.network.TL(self.node(1), self.node(2), beta, L, Z0)
        self.network.TL(self.node(4), self.node(3), beta, L, Z0)
        self.network.TL(self.node(1), self.node(4), beta, L, self.Z0)
        self.network.TL(self.node(3), self.node(3), beta, L, self.Z0)

class RatraceCoupler(FourNodeSubCircuit):

    def __init__(self, Z0: float | SimParam, fc: float | SimParam, er: float | SimParam = 1):
        super().__init__()
        self.Z0 = Z0
        self.fc = fc
        self.er = er
    
    def __on_connect__(self):
        wavelength = 299792458 / (self.fc*np.sqrt(self.er))
        L = wavelength / 4

        def beta(f):
            return 2*np.pi*f*np.sqrt(self.er)/(299792458)
        
        Z0 = self.Z0*np.sqrt(2)

        self.network.TL(self.node(1), self.node(2), beta, L, Z0)
        self.network.TL(self.node(2), self.node(3), beta, L, Z0)
        self.network.TL(self.node(3), self.node(4), beta, L, Z0)
        self.network.TL(self.node(1), self.node(4), beta, 3*L, Z0)

    @property
    def sum(self) -> Node:
        return self.node(2)
    
    @property
    def diff(self) -> Node:
        return self.node(4)
    
    @property
    def delta(self) -> Node:
        return self.node(4)
    
    @property
    def p1(self) -> Node:
        return self.node(1)
    
    @property
    def p2(self) -> Node:
        return self.node(3)
    
class Transformer(TwoNodeSubCircuit):

    def __init__(self, 
                 Z1: float | SimParam,
                 Z2: float | SimParam,
                 fc: float | SimParam, 
                 N_sections: int = 1,
                 ripple: float | SimParam = 0.05,
                 er: float | SimParam = 1):
        super().__init__()
        self.Z1 = Z1
        self.Z2 = Z2
        self.fc = fc
        self.N_sections = N_sections
        self.ripple = ripple
        self.er = er

    def __on_connect__(self):
        impedances = impedance_transformer_cheb(self.Z1, self.Z2, self.ripple, self.N_sections)[1:-1]
        wavelength = 299792458 / (self.fc * np.sqrt(self.er))
        L = wavelength / 4
        nodes = [self.node(1),] + [self.network.node() for _ in range(self.N_sections-1)] + [self.node(2),]
        for n1, n2, z0 in zip(nodes[:-1],nodes[1:],impedances):
            self.network.TL(n1, n2, lambda f: 2*np.pi*f/299792458 * np.sqrt(self.er), L, float(z0))
        


"""
In this temo we will look at how to import a touchstone file.

We use the touchstone file as was given on the Microwave101 website: https://www.microwaves101.com/encyclopedias/snp-format

"""

import heavi as hv
from pathlib import Path

# We generate our Model object
m = hv.Model()

# We get our Ground node
gnd = m.gnd
# We create two ports. This quick_port() method automatically returns the two nodes.

n1 = m.quick_port(50)
n2 = m.quick_port(50)

# Because I don't know where you'll put this Python file and the s2p file in the examples folder, this will always look in the directory that this Python file is in.
step_path = Path(__file__).parent / 'cha3024_99f_lna.s2p' 

# We can simply create our LNA object as a File Based N-port network.
LNA = hv.FileBasedNPort(str(step_path), 2)

# To connect our LNA to the port nodes we sumply call the Connect method.
LNA.connect(n1, n2) # Here we add the nodes that we want to connect our LNA to in order of the ports, port1, port2 etc.

# And we run!
fs = hv.frange(1e9, 20e9, 2001)

S = m.run_sp(fs)

hv.plot_s_parameters(fs, [S.S11, S.S21, S.S12, S.S22], labels=['S11','S21','S12','S22'])